from __future__ import annotations

import functools
import operator
from typing import Callable, Iterable

import awkward as ak
import pyarrow.compute as pc

from akimbo.apply_tree import dec

methods = [
    _ for _ in (dir(ak)) if not _.startswith(("_", "ak_")) and not _[0].isupper()
] + ["apply", "array", "explode", "dt", "str"]

df_methods = sorted(methods + ["merge"])
series_methods = sorted(methods + ["unmerge"])


def radd(left, right):
    return right + left


def rsub(left, right):
    return right - left


def rmul(left, right):
    return right * left


def rdiv(left, right):
    return right / left


def rtruediv(left, right):
    return right / left


def rfloordiv(left, right):
    return right // left


def rmod(left, right):
    # check if right is a string as % is the string
    # formatting operation; this is a TypeError
    # otherwise perform the op
    if isinstance(right, str):
        typ = type(left).__name__
        raise TypeError(f"{typ} cannot perform the operation mod")

    return right % left


def rdivmod(left, right):
    return divmod(right, left)


def rpow(left, right):
    return right**left


def rand_(left, right):
    return operator.and_(right, left)


def ror_(left, right):
    return operator.or_(right, left)


def rxor(left, right):
    return operator.xor(right, left)


class AbstractMethodError(NotImplementedError):
    pass


class ArithmeticMixin:
    @classmethod
    def _create_op(cls, op):
        raise AbstractMethodError(cls)

    @classmethod
    def _create_op(cls, op):
        raise AbstractMethodError(cls)

    @classmethod
    def _create_op(cls, op):
        raise AbstractMethodError(cls)

    @classmethod
    def _add_arithmetic_ops(cls) -> None:
        setattr(cls, "__add__", cls._create_op(operator.add))
        setattr(cls, "__radd__", cls._create_op(radd))
        setattr(cls, "__sub__", cls._create_op(operator.sub))
        setattr(cls, "__rsub__", cls._create_op(rsub))
        setattr(cls, "__mul__", cls._create_op(operator.mul))
        setattr(cls, "__rmul__", cls._create_op(rmul))
        setattr(cls, "__pow__", cls._create_op(operator.pow))
        setattr(cls, "__rpow__", cls._create_op(rpow))
        setattr(cls, "__mod__", cls._create_op(operator.mod))
        setattr(cls, "__rmod__", cls._create_op(rmod))
        setattr(cls, "__floordiv__", cls._create_op(operator.floordiv))
        setattr(cls, "__rfloordiv__", cls._create_op(rfloordiv))
        setattr(cls, "__truediv__", cls._create_op(operator.truediv))
        setattr(cls, "__rtruediv__", cls._create_op(rtruediv))
        setattr(cls, "__divmod__", cls._create_op(divmod))
        setattr(cls, "__rdivmod__", cls._create_op(rdivmod))

    @classmethod
    def _add_comparison_ops(cls) -> None:
        setattr(cls, "__eq__", cls._create_op(operator.eq))
        setattr(cls, "__ne__", cls._create_op(operator.ne))
        setattr(cls, "__lt__", cls._create_op(operator.lt))
        setattr(cls, "__gt__", cls._create_op(operator.gt))
        setattr(cls, "__le__", cls._create_op(operator.le))
        setattr(cls, "__ge__", cls._create_op(operator.ge))

    @classmethod
    def _add_logical_ops(cls) -> None:
        setattr(cls, "__and__", cls._create_op(operator.and_))
        setattr(cls, "__rand__", cls._create_op(rand_))
        setattr(cls, "__or__", cls._create_op(operator.or_))
        setattr(cls, "__ror__", cls._create_op(ror_))
        setattr(cls, "__xor__", cls._create_op(operator.xor))
        setattr(cls, "__rxor__", cls._create_op(rxor))

    @classmethod
    def _add_all(cls):
        cls._add_logical_ops()
        cls._add_arithmetic_ops()
        cls._add_comparison_ops()


class Accessor(ArithmeticMixin):
    """Bring the awkward API to dataframes and series"""

    aggregations = True  # False means data is partitioned
    series_type = ()
    dataframe_type = ()
    subaccessors = {}

    def __init__(self, obj, behavior=None):
        self._obj = obj
        self._behavior = behavior

    def __call__(self, *args, behavior=None, **kwargs):
        return Accessor(self._obj, behavior=behavior)

    cast = dec(pc.cast)

    @property
    def accessor(self):
        # if we use `dec`, which expects to work on
        return self

    @classmethod
    def is_series(cls, data):
        return isinstance(data, cls.series_type)

    @classmethod
    def is_dataframe(cls, data):
        return isinstance(data, cls.dataframe_type)

    @classmethod
    def _to_output(cls, data):
        # TODO: clarify protocol here; can data be in arrow already?
        raise NotImplementedError

    def to_output(self, data=None):
        data = data if data is not None else self.array
        if not isinstance(data, Iterable):
            return data
        return self._to_output(data)

    def apply(self, fn: Callable, where=None, **kwargs):
        """Perform arbitrary function on all the values of the series

        The function should take an ak array as input and produce an
        ak array or scalar.
        """
        if where:
            bits = tuple(where.split("."))
            arr = self.array
            part = arr.__getitem__(bits)
            out = fn(part, **kwargs)
            final = ak.with_field(arr, out, where=where)
        else:
            final = fn(self.array)
        return self.to_output(final)

    def __getitem__(self, item):
        out = self.array.__getitem__(item)
        return self.to_output(out)

    def __dir__(self) -> Iterable[str]:
        attrs = (_ for _ in dir(self.array) if not _.startswith("_"))
        meths = series_methods if self.is_series(self._obj) else df_methods
        return sorted(set(attrs) | set(meths))

    def with_behavior(self, behavior, where=()):
        """Assign a behavior to this array-of-records"""
        # TODO: compare usage with sub-accessors
        # TODO: implement where= (assign directly to ._paraneters["__record__"]
        #  of output's layout. In this case, behaviour is a dict of locations to apply to.
        #  and we can continually add to it (or accept a dict)
        # beh = self._behavior.copy()
        # if isinstance(behavior, dict):
        #    beh.update(behavior)
        # else:
        #    # str or type
        #    beh[where] = behaviour
        return type(self)(self._obj, behavior)

    with_name = with_behavior  # alias - this is the upstream name

    asa = with_behavior  # favoured name

    def __array_function__(self, *args, **kwargs):
        return self.array.__array_function__(*args, **kwargs)

    def __array_ufunc__(self, *args, **kwargs):
        if args[1] == "__call__":
            return self.to_output(args[0](self.array, *args[3:], **kwargs))
        raise NotImplementedError

    @property
    def arrow(self) -> ak.Array:
        """Data as an arrow array"""
        return self.to_arrow(self._obj)

    @classmethod
    def to_arrow(cls, data):
        """Data as an arrow array"""
        raise NotImplementedError

    @property
    def array(self) -> ak.Array:
        """Data as an awkward array"""
        if self._behavior:
            return ak.with_name(ak.from_arrow(self.arrow), self._behavior)
        return ak.from_arrow(self.arrow)

    @classmethod
    def register_accessor(cls, name, klass):
        # TODO: check clobber?
        cls.subaccessors[name] = klass

    def rename(self, where, to):
        """Assign new field name to given location in the structure

        Parameters
        ----------
        where: str | tuple[str]
            location we would like to rename
        to: str
            new name
        """
        arr = self.array
        lay = ak.copy(arr.layout)
        where = list(where) if isinstance(where, tuple) else [where]
        parent = None
        bit = lay
        while where:
            if getattr(bit, "contents", None):
                this = bit.fields.index(where.pop(0))
                parent, bit = bit, bit.contents[this]
            else:
                parent, bit = bit, bit.content
        parent.fields[this] = to
        return self.to_output(ak.Array(lay))

    def merge(self):
        """Make a single complex series out of the columns of a dataframe"""
        if not self.is_dataframe(self._obj):
            raise ValueError("Can only merge on a dataframe")
        out = {}
        for k in self._obj.columns:
            # TODO: partial merge when column names are like "record.field"
            out[k] = self._obj[k].ak.array
        arr = ak.Array(out)
        return self.to_output(arr)

    def unmerge(self):
        """Make dataframe out of a series of record type"""
        arr = self.array
        if not arr.fields:
            raise ValueError("Not array-of-records")
        # TODO: partial unmerge when (some) fields are records
        out = {k: self.to_output(arr[k]) for k in arr.fields}
        return self.dataframe_type(out)

    def join(
        self,
        other,
        key: str,
        colname: str = "match",
        sort: bool = False,
        rkey: str | None = None,
        numba: bool = True,
    ):
        """DB ORM-style left join to other dataframe/series with nesting but no copy

        Related records of the ``other`` table will appear as a list under the new field
        ``colname`` for all matching keys. This is the speed and memory efficient way
        to doing a pandas-style merge/join, which explodes out the values to a much
        bigger memory footprint.

        Parameters
        ----------
        other: series or table
        key: name of the field in this table to match on
        colname: the field that will be added to each record. This field will exist even
            if there are no matches, in which case the list will be empty.
        sort: if False, assumes that they key is sorted in both tables. If True, an
            argsort is performed first, and the match is done by indexing. This may be
            significantly slower.
        rkey: if the name of the field to match on is different in the ``other`` table.
        numba: the matching algorithm will go much faster using numba. However, you can
            set this to False if you do not have numba installed.
        """
        from akimbo.io import join

        out = join(
            self.array, other.ak.array, key, colname=colname, sort=sort, rkey=rkey
        )
        return self.to_output(out)

    @classmethod
    def _create_op(cls, op):
        """Make functions to perform all the arithmetic, logical and comparison ops"""

        def run(self, *args, **kwargs):
            ar2 = (ar.ak.array if hasattr(ar, "ak") else ar for ar in args)
            ar3 = (ar.array if isinstance(ar, cls) else ar for ar in ar2)
            return self.to_output(op(self.array, *ar3, **kwargs))

        return run

    def __getattr__(self, item):
        arr = self.array
        if hasattr(arr, item) and callable(getattr(arr, item)):
            func = getattr(arr, item)
            args = ()
        elif item in self.subaccessors:
            return self.subaccessors[item](self)
        elif hasattr(ak, item):
            func = getattr(ak, item)
            args = (arr,)
        else:
            raise AttributeError(item)

        @functools.wraps(func)
        def f(*others, **kwargs):
            others = [
                other.ak.array
                if isinstance(other, (self.series_type, self.dataframe_type))
                else other
                for other in others
            ]
            kwargs = {
                k: v.ak.array
                if isinstance(v, (self.series_type, self.dataframe_type))
                else v
                for k, v in kwargs.items()
            }

            ak_arr = func(*args, *others, **kwargs)
            if isinstance(ak_arr, ak.Array):
                return self.to_output(ak_arr)
            return ak_arr

        return f

    def __init_subclass__(cls, **kwargs):
        # auto add methods to all derivative classes
        cls._add_all()



from __future__ import annotations

import awkward as ak
import fsspec
import numpy as np


def ak_to_series(ds, backend="pandas", extract=True):
    if backend == "pandas":
        import akimbo.pandas

        s = akimbo.pandas.PandasAwkwardAccessor._to_output(ds)
    elif backend == "polars":
        import akimbo.polars

        s = akimbo.polars.PolarsAwkwardAccessor._to_output(ds)
    elif backend == "dask":
        import akimbo.dask

        # TODO: actually don't use this, use dask-awkward, or dask.dataframe
        s = akimbo.polars.PolarsAwkwardAccessor._to_output(ds)
    elif backend == "cudf":
        import akimbo.cudf

        s = akimbo.cudf.CudfAwkwardAccessor._to_output(ds)
    else:
        raise ValueError("Backend must be in {'pandas', 'polars', 'dask'}")
    if extract and ds.fields:
        return s.ak.unmerge()
    return s


def read_parquet(
    url: str,
    storage_options: dict | None = None,
    extract: bool = True,
    backend: str = "pandas",
    **kwargs,
):
    """Read a Parquet dataset with nested data into a Series or DataFrame.

    This may cope with some deeply nested structures that pandas refuses
    to read by itself.

    You can pass a selection of columns to read (list of strings), and
    other columns will not be parsed into memory. Each of these labels
    may be a root of deeper-nested structs, or use "*" globbing.

    Parameters
    ----------
    url: data location
        Directory with data files, single file or glob pattern
    storage_options: any arguments for an fsspec backend
    extract: whether to turn top-level records into a dataframe. If False,
        will return a series.
    backend: one of "pandas", "polars" or "dask"
    """
    # TODO: is this useful compared to pyarrow.parquet? Describe differences
    ds = ak.from_parquet(url, storage_options=storage_options, **kwargs)
    return ak_to_series(ds, backend, extract=extract)


def read_json(
    url: str,
    storage_options: dict | None = None,
    schema: dict | None = None,
    extract: bool = True,
    backend: str = "pandas",
    **kwargs,
):
    """Read a JSON dataset with nested data into a Series or DataFrame.

    You can pass a selection of columns to read (list or jsonschema format),
    using ``schema=``, and
    other columns will not be parsed into memory. See the docs for
    ak.from_json for further details.

    (examples to come)

    Parameters
    ----------
    url: data location (may include glob characters)
    storage_options: any arguments for an fsspec backend
    schema: if given, the JSONschema expected in the data; this allows for
        selecting only some part of the record structure, this saving on
        some parsing time and potentially a lot of memory footprint. Even if reading
        all the data, providing a schema will lead to better performance.
    extract: whether to turn top-level records into a dataframe. If False,
        will return a series.
    backend: one of "pandas", "polars" or "dask"
    """
    # TODO: implement columns=["field1", "field2.sub", ...] style schema
    #  using dak.lib.io.layout_to_jsonschema
    with fsspec.open_files(url, **(storage_options or {})) as f:
        ds = ak.concatenate(
            [ak.from_json(_, line_delimited=True, schema=schema, **kwargs) for _ in f]
        )
    return ak_to_series(ds, backend, extract=extract)


def get_json_schema(
    url: str, storage_options: dict | None = None, nbytes: int = 1_000_000, **kwargs
):
    """Get JSONSchema representation of the contents of a line-delimited JSON file

    Currently, requires dask_awkward to be installed, which in turn required dask

    Parameters
    ----------
    url: file location
    storage_options: passed to fsspec
    nbytes: how much of the file to read in infer the types. Must be at least one line, and
        should be representative of all the data.

    Returns
    -------
    JSONschema dictionary
    """
    from dask_awkward.lib.io.json import layout_to_jsonschema

    with fsspec.open(url, **(storage_options or {})) as f:
        data = f.read(nbytes).rsplit(b"\n", 1)[0]
    arr = ak.from_json(data, line_delimited=True, **kwargs)
    return layout_to_jsonschema(arr.layout)


def read_avro(
    url: str,
    storage_options: dict | None = None,
    extract: bool = True,
    backend: str = "pandas",
    **kwargs,
):
    """Read AVRO structured data files

    Parameters
    ----------
    url: data location (may include glob characters)
    storage_options: any arguments for an fsspec backend
    extract: whether to turn top-level records into a dataframe. If False,
        will return a series.
    backend: one of "pandas", "polars" or "dask"
    """
    from awkward._connect.avro import ReadAvroFT
    from awkward.operations.ak_from_avro_file import _impl

    with fsspec.open(url, **(storage_options or {})) as f:
        # TODO: ak.from_avro_file broken with file-like
        reader = ReadAvroFT(f, limit_entries=None, debug_forth=False)
        ds = _impl(*reader.outcontents, highlevel=True, attrs=None, behavior=None)
    return ak_to_series(ds, backend, extract=extract)


def get_avro_schema(
    url: str,
    storage_options: dict | None = None,
):
    """Fetch ak form of the schema defined in given avro file"""
    from awkward._connect.avro import ReadAvroFT

    with fsspec.open(url, "rb", **(storage_options or {})) as f:
        reader = ReadAvroFT(f, limit_entries=1, debug_forth=False)
        form, length, container = reader.outcontents
    return form


_jitted = [None]


def join(
    table1: ak.Array,
    table2: ak.Array,
    key: str,
    colname: str = "match",
    sort: bool = False,
    rkey: str = None,
    numba=True,
):
    """Make nested ORM-style left join on common key in two tables

    Assuming ``key`` is a field in each table, the output will look like
    ``table1`` but with an extra column ``colname`` containing a list of
    records from matching rows in ``table2``.
    """
    rkey = rkey or key
    # assert key fields are 1D? allow optional?
    if sort:
        # indexed view is not cache friendly; real sort is better but copies
        table1 = table1[ak.argsort(table1[key], axis=0)]
        table2 = table2[ak.argsort(table2[rkey], axis=0)]
    if numba:
        if _jitted[0] is None:
            try:
                from numba import njit

                # per-session cache, cache=True doesn't work
                _jitted[0] = njit()(_merge)
            except ImportError:
                raise ImportError(
                    "numba is required for fast joins, but you can choose to run with"
                    " numba=False"
                )
        merge = _jitted[0]
    else:
        merge = _merge

    counts = np.empty(len(table1), dtype="uint64")
    # TODO: the line below over-allocates, can switch to somehing growable
    matches = np.empty(len(table2), dtype="uint64")
    # TODO: to_numpy(allow_missong) makes this a bit faster, but is not
    #  not GPU general
    counts, matches, ind = merge(table1[key], table2[key], counts, matches)
    matches.resize(int(ind), refcheck=False)
    indexed = table2[matches]
    listy = ak.unflatten(indexed, counts)
    return ak.with_field(table1, listy, colname)


def _merge(ind1, ind2, counts, matches):
    len2 = len(ind2)
    j = 0
    offind = 0
    matchind = 0
    last = 0
    for i in ind1:
        while True:
            if j >= len2:
                break
            if i > ind2[j]:
                # ID not yet found
                j += 1
                continue
            if i < ind2[j]:
                # no more entrie
                break
            # hit
            while True:
                matches[matchind] = j
                j += 1
                matchind += 1
                if j >= len2 or i != ind2[j]:
                    break
        counts[offind] = matchind - last
        last = matchind
        offind += 1
    return counts, matches, matchind



from __future__ import annotations

import functools
from collections.abc import Callable

import awkward as ak
import pyarrow.compute as pc

from akimbo.apply_tree import dec
from akimbo.mixin import Accessor


def match_string(*layout):
    return layout[0].is_list and layout[0].parameter("__array__") == "string"


def _encode(layout):
    layout._parameters["__array__"] = "bytestring"
    layout.content._parameters["__array__"] = "byte"
    return layout


def match_bytestring(*layout):
    return layout[0].is_list and layout[0].parameter("__array__") == "bytestring"


def _decode(layout):
    layout._parameters["__array__"] = "string"
    layout.content._parameters["__array__"] = "char"
    return layout


_SA_METHODMAPPING = {
    "endswith": "ends_with",
    "isalnum": "is_alnum",
    "isalpha": "is_alpha",
    "isascii": "is_ascii",
    "isdecimal": "is_decimal",
    "isdigit": "is_digit",
    "islower": "is_lower",
    "isnumeric": "is_numeric",
    "isprintable": "is_printable",
    "isspace": "is_space",
    "istitle": "is_title",
    "isupper": "is_upper",
    "startswith": "starts_with",
}
methods = [
    aname
    for aname in (dir(ak.str))
    if not aname.startswith(("_", "akstr_")) and not aname[0].isupper()
]

# make sensible defaults for strptime
strptime = functools.wraps(pc.strptime)(
    lambda *args, format="%FT%T", unit="s", error_is_null=True, **kw:
        pc.strptime(*args, format=format, unit=unit, error_is_null=error_is_null)
)


class StringAccessor:
    """String operations on nested/var-length data"""

    def __init__(self, accessor):
        self.accessor = accessor

    _encode_f = dec(_encode, match=match_string, inmode="ak")

    def encode(self, encoding: str = "utf-8"):
        """Encode Series of strings to Series of bytes. Leaves non-strings alone."""
        if encoding.lower() not in ["utf-8", "utf8"]:
            raise NotImplementedError
        return self._encode_f()

    _decode_f = dec(_decode, match=match_bytestring, inmode="ak")

    def decode(self, encoding: str = "utf-8"):
        """Decode Series of bytes to Series of strings. Leaves non-bytestrings alone.

        Validity of UTF8 is *not* checked.
        """
        if encoding.lower() not in ["utf-8", "utf8"]:
            raise NotImplementedError
        return self._decode_f()

    @staticmethod
    def method_name(attr: str) -> str:
        return _SA_METHODMAPPING.get(attr, attr)

    def __getattr__(self, attr: str) -> Callable:
        attr = self.method_name(attr)
        fn = getattr(ak.str, attr)

        @functools.wraps(fn)
        def f(*args, **kwargs):
            arr = fn(self.accessor.array, *args, **kwargs)
            # idx = self.accessor._obj.index
            if isinstance(arr, ak.Array):
                return self.accessor.to_output(arr)
            return arr

        return f

    strptime = dec(strptime, match=match_string)

    def __dir__(self) -> list[str]:
        return sorted(methods + ["strptime"])


Accessor.register_accessor("str", StringAccessor)



class NoAttributes:
    """Allows importing akimbo.cudf even if cudf isn't installed

    This is done so that sphinx can still build docs on non-GPU systems.
    """

    def __dir__(self):
        return []

    def __getattr__(self, item):
        if item == "__qualname__":
            return "akimbo.utils.DummyAttributesObject"
        if item == "__type_params__":
            return ()
        return self

    def __call__(self, *args, **kwargs):
        return self

    __name__ = "DummyAttributesObject"
    __doc__ = None
    __annotations__ = None



import awkward as ak
import polars as pl

from akimbo.mixin import Accessor


@pl.api.register_series_namespace("ak")
@pl.api.register_dataframe_namespace("ak")
class PolarsAwkwardAccessor(Accessor):
    """Perform awkward operations on a polars series or dataframe"""

    series_type = pl.Series
    dataframe_type = pl.DataFrame

    @classmethod
    def _to_output(cls, arr):
        return pl.from_arrow(ak.to_arrow(arr, extensionarray=False))

    @classmethod
    def to_arrow(cls, data):
        return data.to_arrow()



from __future__ import annotations

import functools
import inspect
from typing import Callable, Literal, Sequence

import awkward as ak
import pyarrow as pa


def leaf(*layout, **_):
    """True for the lowest elements of any akwward layout tree"""
    return layout[0].is_leaf


def run_with_transform(
    arr: ak.Array,
    op,
    match=leaf,
    outtype=None,
    inmode="arrow",
    others=(),
    match_kwargs=None,
    **kw,
) -> ak.Array:
    def func(layout, **kwargs):
        if not isinstance(layout, tuple):
            layout = (layout,)
        if all(match(lay, **(match_kwargs or {})) for lay in layout):
            if inmode == "arrow":
                out = ak.str._apply_through_arrow(
                    op, *layout, **kw, **(match_kwargs or {})
                )
            elif inmode == "numpy":
                # works on numpy/cupy contents
                out = op(*(lay.data for lay in layout), **kw, **(match_kwargs or {}))
            else:
                out = op(*layout, **kw, **(match_kwargs or {}))
            return outtype(out) if callable(outtype) else out

    return ak.transform(func, arr, *others)


def dec(
    func: callable,
    match: Callable[[ak.contents.Content], bool] = leaf,
    outtype: Callable[[ak.contents.Content], ak.contents.Content] | None = None,
    inmode: Literal["arrow", "numpy", "ak"] = "arrow",
):
    """Make a nested/ragged version of an operation to apply throughout a tree

    Parameters
    ----------
    func: which we want to apply to (parts of) inputted data
    match: function to determine if a part of the data structure matches the type we want to
        operate on
    outtype: postprocessing function after transform
    inmode: how ``func`` expects its inputs: as awkward arrays (ak), numpy or arrow
    """

    @functools.wraps(func)
    def f(self, *args, where=None, match_kwargs=None, **kwargs):
        if not (
            where is None
            or isinstance(where, str)
            or isinstance(where, Sequence)
            and all(isinstance(_, str) for _ in where)
        ):
            raise ValueError
        others = []
        if args:
            sig = list(inspect.signature(func).parameters)[1:]
            for k, arg in zip(sig, args):
                if isinstance(arg, ak.Array):
                    others.append(arg)
                elif isinstance(
                    arg, (self.accessor.series_type, self.accessor.dataframe_type)
                ):
                    others.append(arg.ak.array)
                elif isinstance(arg, pa.Array):
                    others.append(ak.from_arrow(arg))
                else:
                    kwargs.update({k: arg for k, arg in zip(sig, args)})
        if where:
            bits = tuple(where.split("."))
            arr = self.accessor.array
            part = arr.__getitem__(bits)
            others = [o.__getitem__(bits) for o in others]
            out = run_with_transform(
                part,
                func,
                match=match,
                outtype=outtype,
                inmode=inmode,
                others=others,
                match_kwargs=match_kwargs,
                **kwargs,
            )
            final = ak.with_field(arr, out, where=where)
            return self.accessor.to_output(final)
        else:
            return self.accessor.to_output(
                run_with_transform(
                    self.accessor.array,
                    func,
                    match=match,
                    outtype=outtype,
                    inmode=inmode,
                    others=others,
                    match_kwargs=match_kwargs,
                    **kwargs,
                )
            )

    f.__doc__ = f"""Run vectorized functions on nested/ragged/complex array

where: None | str | Sequence[str, ...]
    if None, will attempt to apply the kernel throughout the nested structure,
    wherever correct types are encountered. If where is given, only the selected
    part of the structure will be considered, but the output will retain
    the original shape. A fieldname or sequence of fieldnames to descend into
    the tree are acceptable
match_kwargs: None | dict
    any extra field identifiers for matching a record as OK to process

{'--Kernel documentation follows from the original function--' if f.__doc__ else ''}

{f.__doc__ or str(f)}
"""

    return f



import functools

import awkward as ak
import pyarrow.compute as pc

from akimbo.apply_tree import dec
from akimbo.mixin import Accessor


def match(*layouts):
    return layouts[0].is_leaf and layouts[0].dtype.kind == "M"


dec_t = functools.partial(dec, match=match)


class DatetimeAccessor:
    def __init__(self, accessor) -> None:
        self.accessor = accessor

    # listed below https://arrow.apache.org/docs/python/generated/
    # pyarrow.compute.ceil_temporal.html
    ceil_temporal = dec_t(pc.ceil_temporal)
    floor_temporal = dec_t(pc.floor_temporal)
    reound_temporal = dec_t(pc.round_temporal)
    strftime = dec_t(pc.strftime)
    strptime = dec_t(pc.strptime)
    day = dec_t(pc.day)
    day_of_week = dec_t(pc.day_of_week)
    day_of_year = dec_t(pc.day_of_year)
    hour = dec_t(pc.hour)
    iso_week = dec_t(pc.iso_week)
    iso_year = dec_t(pc.iso_year)
    iso_calendar = dec_t(pc.iso_calendar)
    is_leap_year = dec_t(pc.is_leap_year)
    microsecond = dec_t(pc.microsecond)
    millisecond = dec_t(pc.millisecond)
    minute = dec_t(pc.minute)
    month = dec_t(pc.month)
    nanosecond = dec_t(pc.nanosecond)
    quarter = dec_t(pc.quarter)
    second = dec_t(pc.second)
    subsecond = dec_t(pc.subsecond)
    us_week = dec_t(pc.us_week)
    us_year = dec_t(pc.us_year)
    week = dec_t(pc.week)
    year = dec_t(pc.year)
    year_month_day = dec_t(pc.year_month_day)

    day_time_interval_between = dec_t(pc.day_time_interval_between)
    days_between = dec_t(pc.days_between)
    hours_between = dec_t(pc.hours_between)
    microseconds_between = dec_t(pc.microseconds_between)
    milliseconds_between = dec_t(pc.milliseconds_between)
    minutes_between = dec_t(pc.minutes_between)
    month_day_nano_interval_between = dec_t(pc.month_day_nano_interval_between)
    month_interval_between = dec_t(pc.month_interval_between)
    nanoseconds_between = dec_t(pc.nanoseconds_between)
    quarters_between = dec_t(pc.quarters_between)
    seconds_between = dec_t(pc.seconds_between)
    weeks_between = dec_t(pc.weeks_between)
    years_between = dec_t(pc.years_between)

    strftime = dec_t(pc.strftime)

    # TODO: timezone conversion


def _to_arrow(array):
    array = _make_unit_compatible(array)
    return ak.to_arrow(array, extensionarray=False)


def _make_unit_compatible(array):
    # TODO, actually convert units if not compatible
    return array


Accessor.register_accessor("dt", DatetimeAccessor)



import functools
from typing import Callable

import awkward as ak

from akimbo.utils import NoAttributes

try:
    import cudf
    from cudf import DataFrame, Series
    from cudf import _lib as libcudf
    from cudf.core.column.datetime import DatetimeColumn
    from cudf.core.column.string import StringMethods
except ImportError:
    StringMethods = NoAttributes()
    DatetimeColumn = NoAttributes()
    libcudf = NoAttributes()
    DataFrame = Series = NoAttributes()


from akimbo.ak_from_cudf import cudf_to_awkward as from_cudf
from akimbo.apply_tree import dec, leaf
from akimbo.datetimes import DatetimeAccessor
from akimbo.datetimes import match as match_t
from akimbo.mixin import Accessor
from akimbo.strings import StringAccessor


def match_string(arr):
    return arr.parameters.get("__array__", "") == "string"


class CudfStringAccessor(StringAccessor):
    """String operations on nested/var-length data"""

    def decode(self, encoding: str = "utf-8"):
        raise NotImplementedError(
            "cudf does not support bytearray type, so we can't automatically identify them"
        )

    def encode(self, encoding: str = "utf-8"):
        raise NotImplementedError("cudf does not support bytearray type")


def dec_cu(op, match=match_string):
    @functools.wraps(op)
    def f(lay, **kwargs):
        # op(column, ...)->column
        col = op(lay._to_cudf(cudf, None, len(lay)), **kwargs)
        return from_cudf(cudf.Series(col)).layout

    return dec(func=f, match=match, inmode="ak")


for meth in dir(StringMethods):
    if meth.startswith("_"):
        continue

    @functools.wraps(getattr(StringMethods, meth))
    def f(lay, method=meth, **kwargs):
        # this is different from dec_cu, because we need to instantiate StringMethods
        # before getting the method from it
        col = getattr(
            StringMethods(cudf.Series(lay._to_cudf(cudf, None, len(lay)))), method
        )(**kwargs)
        return from_cudf(col).layout

    setattr(CudfStringAccessor, meth, dec(func=f, match=match_string, inmode="ak"))


class CudfDatetimeAccessor(DatetimeAccessor):
    ...


for meth in dir(DatetimeColumn):
    if meth.startswith("_") or meth == "strptime":
        # strptime belongs in .str, not here!
        continue

    @functools.wraps(getattr(DatetimeColumn, meth))
    def f(lay, method=meth, **kwargs):
        # this is different from dec_cu, because we need to instantiate StringMethods
        # before getting the method from it
        m = getattr(lay._to_cudf(cudf, None, len(lay)), method)
        if callable(m):
            col = m(**kwargs)
        else:
            # attributes giving components
            col = m
        return from_cudf(cudf.Series(col)).layout

    if isinstance(getattr(DatetimeColumn, meth), property):
        setattr(
            CudfDatetimeAccessor,
            meth,
            property(dec(func=f, match=match_t, inmode="ak")),
        )
    else:
        setattr(CudfDatetimeAccessor, meth, dec(func=f, match=match_t, inmode="ak"))


class CudfAwkwardAccessor(Accessor):
    series_type = Series
    dataframe_type = DataFrame

    @classmethod
    def _to_output(cls, arr):
        if isinstance(arr, ak.Array):
            return ak.to_cudf(arr)
        elif isinstance(arr, ak.contents.Content):
            return arr._to_cudf(cudf, None, len(arr))
        return arr

    @classmethod
    def to_array(cls, data) -> ak.Array:
        return from_cudf(data)

    @property
    def array(self) -> ak.Array:
        return self.to_array(self._obj)

    @property
    def str(self):
        """Nested string operations"""
        # need to find string ops within cudf
        return CudfStringAccessor(self)

    cast = dec_cu(libcudf.unary.cast, match=leaf)

    @property
    def dt(self):
        """Nested datetime operations"""
        # need to find datetime ops within cudf
        return CudfDatetimeAccessor(self)

    def apply(self, fn: Callable, *args, **kwargs):
        if "CPUDispatcher" in str(fn):
            # auto wrap original function for GPU
            raise NotImplementedError
        super().apply(fn, *args, **kwargs)


@property  # type:ignore
def ak_property(self):
    return CudfAwkwardAccessor(self)


Series.ak = ak_property  # no official register function?



import functools
from typing import Iterable

import awkward as ak
import dask.dataframe as dd
from dask.dataframe.extensions import (
    register_dataframe_accessor,
    register_series_accessor,
)

from akimbo.mixin import Accessor as AkAccessor
from akimbo.mixin import df_methods, series_methods
from akimbo.pandas import PandasAwkwardAccessor


class DaskAwkwardAccessor(AkAccessor):
    """Perform awkward operations on a dask series or frame

    These operations are lazy, because of how dask works. Note
    that we use mapping operations here, so any action on
    axis==0 or 1 will produce results per partition, which
    you must then combine.

    To perform intra-partition operations, we recommend you
    use the ``.to_dask_awkward`` method.

    Correct arrow dtypes will be deduced when the input is
    also arrow, which is now the default for the dask
    "dataframe.dtype_backend" config options.
    """

    series_type = dd.Series
    dataframe_type = dd.DataFrame
    aggregations = False  # you need dask-awkward for that

    @staticmethod
    def _to_tt(data):
        # self._obj._meta.convert_dtypes(dtype_backend="pyarrow")
        data = data._meta if hasattr(data, "_meta") else data
        arr = PandasAwkwardAccessor.to_arrow(data)
        return ak.to_backend(ak.from_arrow(arr), "typetracer")

    def to_dask_awkward(self):
        """Convert to dask-awkard.Array object

        This make a s single complex awkward array type out of one or more columns.
        You would do this, in order to use dask-awkward's more advanced inter-
        partition aggregations and optimisation.
        See https://dask-awkward.readthedocs.io/

        c.f., dask_awkward.to_dataframe
        """
        import dask_awkward as dak

        tt = self._to_tt(self._obj)
        return dak.lib.core.new_array_object(
            self._obj.dask, divisions=self._obj.divisions, name=self._obj._name, meta=tt
        )

    @classmethod
    def _create_op(cls, op):
        def run(self, *args, **kwargs):
            try:
                tt = self._to_tt(self._obj)
                ar = (
                    ak.to_backend(ar) if isinstance(ar, (ak.Array, ak.Record)) else ar
                    for ar in args
                )
                ar = [self._to_tt(ar) if hasattr(ar, "ak") else ar for ar in ar]
                out = op(tt, *ar, **kwargs)
                meta = PandasAwkwardAccessor._to_output(
                    ak.typetracer.length_zero_if_typetracer(out)
                )
            except (ValueError, TypeError):
                meta = None

            def inner(data, _=DaskAwkwardAccessor):
                import akimbo.pandas  # noqa: F401

                ar2 = (ar.ak.array if hasattr(ar, "ak") else ar for ar in args)
                out = op(data.ak.array, *ar2, **kwargs)
                return PandasAwkwardAccessor._to_output(out)

            return self._obj.map_partitions(inner, meta=meta)

        return run

    def __getattr__(self, item):
        if item not in dir(self):
            raise AttributeError
        func = getattr(ak, item, None)

        if func:
            orig = self._obj.head()

            @functools.wraps(func)
            def f(*others, **kwargs):
                def func2(data):
                    import akimbo.pandas  # noqa: F401

                    # data and others are pandas objects here
                    return getattr(data.ak, item)(*others, **kwargs)

                return self._obj.map_partitions(func2, meta=func(orig))

        else:
            raise AttributeError(item)
        return f

    def __dir__(self) -> Iterable[str]:
        attrs = (_ for _ in dir(self._obj._meta.ak.array) if not _.startswith("_"))
        meths = series_methods if self.is_series(self._obj) else df_methods
        return sorted(set(attrs) | set(meths))


register_series_accessor("ak")(DaskAwkwardAccessor)
register_dataframe_accessor("ak")(DaskAwkwardAccessor)



from __future__ import annotations

# re-exports
from awkward import behavior
from awkward import metadata_from_parquet as get_parquet_schema
from awkward import mixin_class, mixin_class_method

import akimbo.datetimes as datetimes
import akimbo.mixin as mixin
import akimbo.strings as strings
from akimbo.io import (
    get_avro_schema,
    get_json_schema,
    join,
    read_avro,
    read_json,
    read_parquet,
)
from akimbo.version import version as __version__  # noqa

__all__ = (
    "datetimes",
    "mixin",
    "join",
    "read_avro",
    "read_parquet",
    "read_json",
    "behavior",
    "mixin_class",
    "mixin_class_method",
    "get_parquet_schema",
    "get_json_schema",
    "get_avro_schema",
    "strings",
)
